#
/*
 *  C compiler intermediate code lister
 */

#include "c1.h"

int aopt_debug;
int bopt_debug;

char    *reg_names[] { "ax", "dx", "si", "di", "cx", "bx", "bp", "sp" };
char    *typetab[] { "int", "char", "float", "double", "struct",
			"rstruct", "long", "unsign"};
char    *qualtab[] {"ptr", "func", "array", "lptr"};

int noRtype = 0;
main(argc, argv)
char *argv[];
{
    if (argc>1 && *argv[1] == '-')
    {
	noRtype=1;
	argc--;
	argv++;
    }
    if (argc>1) freopen(argv[1], "r", stdin);

    getree();
}

char typebuf[30];

char *typestr(i)
int i;
{
    char *q, *p;
    if ((i&TYPE) == i) return (typetab[i]);
    p = typebuf;
    for (q=typetab[TYPE&i]; *p++ = *q++;);
    *--p = ' ';
    p++;
    if ((i & XTYPE) == PTR)
	{for (q=qualtab[0]; *p++ = *q++;); }
    if ((i & XTYPE) == FUNC)
	{for (q=qualtab[1]; *p++ = *q++;); }
    if ((i & XTYPE) == ARRAY)
	{for (q=qualtab[2]; *p++ = *q++;); }
#ifdef INTERMED
    if ((i & XTYPE) == LPTR)
	{for (q=qualtab[3]; *p++ = *q++;); }
#endif
    return (typebuf);
}

label(l)
{
	printf("L%d:\n", l);
}

error(s, p1, p2, p3, p4, p5, p6)
{

	nerror++;
	fprintf(stderr, "%d: ", line);
	fprintf(stderr, s, p1, p2, p3, p4, p5, p6);
	putc('\n', stderr);
}

/*
 * Read in an intermediate file.
 */
#define	STKS	100
getree()
{
	register t, op;
	static char s[9];
	struct swtab *swp;
	double atof();
	char numbuf[64];
	struct tname *np;
	struct xtname *xnp;
	struct ftconst *fp;
	struct lconst *lp;
	struct fasgn *sap;
	int lbl, cond, lbl2, lbl3;

	curbase = funcbase;
	for (;;) {
		op = geti();
		if ((op&0177400) != 0177000) {
			error("Intermediate file error");
			exit(1);
		}
		lbl = 0;
		switch(op =& 0377) {

	case SINIT:
		printf("SINIT\t/%x\n", geti());
		break;

	case EOFC:
		printf ("EOFC\n");
		return;

	case BDATA:
		printf ("BDATA\n");
		if (geti() == 1) {
			t = 0;
			printf(".byte	");
			for (;;)  {
				printf("/%x", geti());
				t++;
				if (geti() != 1)
					break;
				if (t == 9) {
					t = 0;
					printf("\n.byte	");
				} else
					putchar(',');
			}
			if (t != 0)
				putchar('\n');
		}
		break;

	case PROG:
		printf("PROG\n");
		break;

	case DATA:
		printf("DATA\n");
		break;

	case BSS:
		printf("BSS\n");
		break;

	case SYMDEF:
		printf ("SYMDEF\t");
		outname(1);
		break;

	case RETRN:
		/* the rtype is the return type of the function.
		   this value is used by the 8086 optimizer to
		   tell about register use */
		if (noRtype) printf ("RETRN\n");
		else printf("RETRN\t%d\n", geti());
		break;

	case CSPACE:
		printf ("CSPACE\t");
		outname(0);
		printf("\t%d\n", geti());
		break;

	case SSPACE:
		printf ("SSPACE  ");
		printf("SSPACE\t%u.\n", (t=geti()));
		break;

	case EVEN:
		printf("EVEN\n");
		break;

	case SAVE:
		printf("SAVE\n");
		break;

	case SETSTK:
		printf ("SETSTK\t%d\n", geti());
		break;

	case PROFIL:
		printf ("PROFIL\t%d\n", geti());
		break;

	case SNAME:
		printf ("SNAME\t");
		outname(0);
		printf("\t%d\n", geti());
		break;

	case ANAME:
		printf ("ANAME\t");
		outname(0);
		printf("\t%d\n", geti());
		break;

	case RNAME:
		printf ("RNAME\t");
		outname(0);
		printf("\t%s\n", reg_names[geti()]);
		break;

	case SWIT:
		t = geti();
		line = geti();
		printf ("SWIT\t%d\tline %d\n", t, line);
		while(t = geti())
		{
			printf (" SWIT lab %d\tval %d\n", t, geti());
		}
		break;

	case C3BRANCH:		/* for fortran [sic] */
		lbl = geti();
		lbl2 = geti();
		lbl3 = geti();
		printf ("C3BRANCH\t%d\t%d\t%d\n", lbl, lbl2, lbl3);
		goto xpr;

	case CBRANCH:
		lbl = geti();
		cond = geti();
		printf ("CBRANCH\t%d\t%d\tline %d\n", lbl, cond, geti());
		break;

	case EXPR:
	xpr:
		line = geti();
		printf ("EXPR\tline %d\n", line);
		break;

	case NAME:
		t = geti();
		if (t==EXTERN) {
			printf ("EXTERN NAME\t%s\t", typestr(geti()));
			outname(1);
		} else {
			lbl = geti();
			printf ("NAME\t%d\t%s\t%d\n", t, typestr(lbl), geti());
		}
		break;

	case CON:
		t = geti();
		printf ("CON\t%s\t%d\n", typestr(t), geti());
		break;

	case LCON:
		geti();	/* ignore type, assume long */
		t = geti();
		op = geti();
		printf ("LCON\t%d\t%d\n", t, op);
		break;

	case FCON:
		printf ("FCON\t%s\n", typestr(geti()));
		outname(1);
		break;

	case FSEL:
		t = geti();
		op = geti();
		printf ("FSEL\t%d\t%s\t%d\n", t, typestr(op), geti());
		break;

	case STRASG:
		t = geti();
		printf ("STRASG\t%d\t%d\n", t, geti());
		break;

	case NULLOP:
		printf("NULLOP\n");
		break;

	case LABEL:
		label(geti());
		break;

	case NLABEL:
		printf ("NLABEL\t");
		outname(1);
		break;

	case RLABEL:
		printf ("RLABEL\t");
		outname(1);
		break;

	case BRANCH:
		printf ("BRANCH\tL%d\n", geti());
		break;

	case SETREG:
		printf("SETREG\t%d\n", geti());
		break;

	default:
		if (opdope[op]&BINARY) {
			printf ("BINOP\t%s\t%s\n", opntab[op], typestr(geti()));
		} else
			printf ("UNOP\t%s\t%s\n", opntab[op], typestr(geti()));
		break;
	}
	}
}

geti()
{
	register i;

	i = getchar();
	i += getchar()<<8;
	return(i);
}

outname(t)
int t;
{
	register char c;
	while (c = getchar()) {
		putchar(c);
	}
	if (t) putchar ('\n');
}

